import inspect
from contextlib import AbstractContextManager
from typing import (
    Type,
    Callable,
    Any,
    List,
    Dict,
    TypeVar,
    Generic,
    Optional,
    TYPE_CHECKING,
)

from .confz_source import ConfZSources

if TYPE_CHECKING:
    from .confz import ConfZ


class SourceChangeManager(AbstractContextManager):
    """Config sources change context manager, allows to change config sources within a
    controlled context and resets everything afterwards."""

    def __init__(self, config_class: Type["ConfZ"], config_sources: ConfZSources):
        self._config_class = config_class
        self._config_sources = config_sources
        self._backup_instance = None
        self._backup_sources = None

    def __enter__(self):
        self._backup_instance = self._config_class.confz_instance
        self._config_class.confz_instance = None

        self._backup_sources = self._config_class.CONFIG_SOURCES
        self._config_class.CONFIG_SOURCES = self._config_sources

        if self._config_class.listeners is not None:
            for listener in self._config_class.listeners:
                listener.change_enter(self)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._config_class.confz_instance = self._backup_instance
        self._config_class.CONFIG_SOURCES = self._backup_sources

        if self._config_class.listeners is not None:
            for listener in self._config_class.listeners:
                listener.change_exit(self)


T = TypeVar("T")


class Listener(Generic[T]):
    """Listener of config, will add singleton mechanism, aware of config changes."""

    def __init__(self, fn: Callable[[], T], config_classes: List[Type["ConfZ"]]):
        if len(inspect.getfullargspec(fn).args) != 0:
            raise ValueError("Callable should not take any arguments")

        for config_class in config_classes:
            if config_class.listeners is None:
                config_class.listeners = []
            config_class.listeners.append(self)

        self._fn = fn
        self._instance: Optional[T] = None
        self._backup_instances: Dict[SourceChangeManager, T] = {}

    @property
    def is_async(self):
        return inspect.iscoroutinefunction(self._fn)

    def __call__(self):
        if self.is_async:

            async def inner():
                if self._instance is None:
                    self._instance = await self._fn()
                return self._instance

        else:

            def inner():
                if self._instance is None:
                    self._instance = self._fn()
                return self._instance

        return inner()

    def change_enter(self, context):
        self._backup_instances[context] = self._instance
        self._instance = None

    def change_exit(self, context):
        self._instance = self._backup_instances[context]
        del self._backup_instances[context]


def depends_on(*args):
    """Decorator to transform a function into a singleton and register it to a set of
    config classes."""
    if len(args) == 1 and inspect.isfunction(args[0]):
        return Listener(args[0], [])

    def inner(fn: Callable[[], Any]) -> Listener:
        return Listener(fn, list(args))

    return inner



from __future__ import annotations  # for sphinx's autodoc_type_aliases

from contextlib import AbstractContextManager
from typing import ClassVar, List, Optional, Any

from pydantic import BaseModel

from .change import SourceChangeManager
from .confz_source import ConfZSources
from .exceptions import ConfZException
from .loaders import get_loader


def _load_config(config_kwargs: dict, confz_sources: ConfZSources) -> dict:
    config = config_kwargs.copy()
    if isinstance(confz_sources, list):
        for confz_source in confz_sources:
            loader = get_loader(type(confz_source))
            loader.populate_config(config, confz_source)
    else:
        loader = get_loader(type(confz_sources))
        loader.populate_config(config, confz_sources)
    return config


# Metaclass of pydantic.BaseModel is not in __all__, so use type(BaseModel).
# ConfZ will be only class with this meta class.
# Both of these things confuse mypy and pylint, so had to disable multiple times.
class ConfZMetaclass(type(BaseModel)):  # type: ignore
    """ConfZ Meta Class, inheriting from the pydantic `BaseModel` MetaClass."""

    # pylint: disable=no-self-argument,no-member
    def __call__(cls, config_sources: ConfZSources = None, **kwargs):
        """Called every time an instance of any ConfZ object is created. Injects the
        config value population and singleton mechanism."""
        if config_sources is not None:
            config = _load_config(kwargs, config_sources)
            return super().__call__(**config)

        if cls.CONFIG_SOURCES is not None:  # type: ignore
            # pylint: disable=access-member-before-definition
            # pylint: disable=attribute-defined-outside-init
            if len(kwargs) > 0:
                raise ConfZException(
                    'Singleton mechanism enabled ("CONFIG_SOURCES" is defined), so '
                    "keyword arguments are not supported"
                )
            if cls.confz_instance is None:  # type: ignore
                config = _load_config(kwargs, cls.CONFIG_SOURCES)  # type: ignore
                cls.confz_instance = super().__call__(**config)
            return cls.confz_instance

        return super().__call__(**kwargs)


class ConfZ(BaseModel, metaclass=ConfZMetaclass):
    """Base class, parent of every config class. Internally wraps :class:`BaseModel`of
    pydantic and behaves transparent except for two cases:

    - If the constructor gets `config_sources` as kwarg, these sources are used as
      input to enrich the other kwargs.
    - If the class has the class variable `CONFIG_SOURCES` defined, these sources are
      used as input.

    In the latter case, a singleton mechanism is activated, returning the same config
    class instance every time the constructor is called."""

    CONFIG_SOURCES: ClassVar[Optional[ConfZSources]] = None  #: Sources to use as input.

    # type is ClassVar[Optional["ConfZ"]] (pydantic throws error with forward ref)
    confz_instance: ClassVar[Optional[Any]] = None  #: *for internal use only*

    # type is ClassVar[Optional[List["Listener"]]] (same here)
    listeners: ClassVar[Optional[List[Any]]] = None  #: *for internal use only*

    class Config:
        allow_mutation = False

    @classmethod
    def change_config_sources(
        cls, config_sources: ConfZSources
    ) -> AbstractContextManager:
        """Change the `CONFIG_SOURCES` class variable within a controlled context.
        Within this context, the sources will be different and the singleton reset.
        This can be useful in unit tests to temporarily change a configuration.

        :param config_sources: The temporary config sources for within the context.
        :return: Context manager for change of config sources.
        """
        return SourceChangeManager(cls, config_sources)



from abc import ABC, abstractmethod
from typing import Dict, Any

from confz.exceptions import ConfZUpdateException


class Loader(ABC):
    """An abstract base class for all config loaders."""

    @classmethod
    def update_dict_recursively(cls, original_dict: Dict, update_dict: Dict):
        """Updates the original dict with the new data. Similar to `dict.update()`, but
        works with nested dicts.

        :param original_dict: The original dictionary to update in-place.
        :param update_dict: The new data.
        :raises ConfZUpdateException: If dict keys contradict each other.
        """
        for key, value in update_dict.items():
            if isinstance(value, dict) and key in original_dict:
                if not isinstance(original_dict[key], dict):
                    raise ConfZUpdateException(
                        f"Config variables contradict each other: "
                        f"Key '{key}' is both a value and a nested dict."
                    )
                cls.update_dict_recursively(original_dict[key], value)
            else:
                original_dict[key] = value

    @classmethod
    def transform_nested_dicts(
        cls, dict_in: Dict[str, Any], separator: str = "."
    ) -> Dict[str, Any]:
        """Transform dictionaries into nested dictionaries, using a separator in the
        keys as hint.

        :param dict_in: A dictionary with string-keys.
        :param separator: The string used to separate dict keys.
                          Default value will no longer be set in a future release.
        :return: The transformed dictionary, splitting keys at the separator and
            creating a new dictionary out of it.
        :raises ConfZUpdateException: If dict keys contradict each other.
        """
        dict_out: Dict[str, Any] = {}
        for key, value in dict_in.items():
            if separator in key and not key.startswith(separator):
                inner_keys = key.split(separator)
                dict_inner = dict_out
                for idx, inner_key in enumerate(inner_keys):
                    if idx == len(inner_keys) - 1:
                        dict_inner[inner_key] = value
                    else:
                        if inner_key not in dict_inner:
                            dict_inner[inner_key] = {}
                        elif not isinstance(dict_inner[inner_key], dict):
                            raise ConfZUpdateException(
                                f"Config variables contradict each other: Key "
                                f"'{inner_key}' is both a value and a nested dict."
                            )
                        dict_inner = dict_inner[inner_key]
            else:
                dict_out[key] = value

        return dict_out

    @classmethod
    @abstractmethod
    def populate_config(cls, config: dict, confz_source):
        """Populate the config-dict with new config arguments based on the source.

        :param config: Config dictionary, gets extended with new arguments
        :param confz_source: Source configuration.
        """



from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional, Union, List, Dict, Any


@dataclass
class ConfZSource:
    """Source configuration for :class:`~confz.ConfZ` models."""


ConfZSources = Union[ConfZSource, List[ConfZSource]]


class FileFormat(Enum):
    """Enum for file format."""

    JSON = "json"  #: JSON file format
    YAML = "yaml"  #: YAML file format


@dataclass
class ConfZFileSource(ConfZSource):
    """Source config for files."""

    file: Optional[Path] = None
    """Specify a config file directly by a path."""
    file_from_env: Optional[str] = None
    """Alternatively, use this environment variable to get the file."""
    file_from_cl: Optional[Union[int, str]] = None
    """Alternatively, use this command line argument to get the file name/path. It can
    be a specific position (integer, e.g. `1`) or after a specific option (string,
    e.g. `\\-\\-config-file`). In the latter case, the file name must follow after
    whitespace, an equal sign between argument and value is not supported right now."""
    folder: Optional[Path] = None
    """The file specified above can optionally be relative to this folder."""
    format: Optional[FileFormat] = None
    """The format of the config file. If not specified, it will be inferred from the
    file ending."""
    encoding: str = "utf-8"
    """The encoding of the file. Default is UTF-8."""


@dataclass
class ConfZEnvSource(ConfZSource):
    """Source config for environment variables and .env files. On loading of the
    source, the dotenv file values (if available) are merged with the environment,
    with environment always taking precedence in case of name collusion. All loaded
    variable names are transformed to lowercase and all dashes are replaced by
    underscores. The definitions below are not case-sensitive and can be written with
    underscore or dash. An exception is `prefix`, which needs to match exactly.
    Dot-notation can be used to access nested configurations."""

    allow_all: bool = False
    """Allow potentially all environment variables to be read as config option."""
    allow: Optional[List[str]] = None
    """Only allow a list of environment variables as input."""
    deny: Optional[List[str]] = None
    """Do not allow to read from environment variables in this list. Useful if
    `allow_all` is set and certain variables should be excluded."""
    prefix: Optional[str] = None
    """The selection above can be narrowed down to a specific prefix, e.g. `CONFIG_`.
    The variables in the lists above or the map below do not need to include this
    prefix, it is automatically added. This option is especially recommended,
    if ´allow_all´ is set."""
    remap: Optional[Dict[str, str]] = None
    """Certain environment variables can be mapped to config arguments with a different
    name."""
    file: Optional[Path] = None
    """Built in .env file loading with lower than environment precedence. Uses UTF-8
    for decoding."""
    nested_separator: str = "."
    """Separator will be used in nested environment variables."""


@dataclass
class ConfZCLArgSource(ConfZSource):
    """Source config for command line arguments. Command line arguments are
    case-sensitive. Dot-notation can be used to access nested configurations. Only
    command line arguments starting with two dashes (\\-\\-) are considered. Between
    argument and value must be whitespace, an equal sign is not supported at the
    moment."""

    prefix: Optional[str] = None
    """Optionally, all command line arguments can have a prefix, e.g. `config_`. The
    prefix does not need to include the two dashes at the beginning. The map below
    does not need to include the prefix, it is automatically added."""
    remap: Optional[Dict[str, str]] = None
    """Certain command line arguments can be mapped to config arguments with a different
    name. The map does not need to include the two dashes at the beginning."""
    nested_separator: str = "."
    """Separator will be used in nested command line arguments."""


@dataclass
class ConfZDataSource(ConfZSource):
    """Source config for raw data, i.e. constants. This can be useful for unit-test
    together with :meth:`~confz.ConfZ.change_config_sources` to inject test data into
    the config."""

    data: Dict[str, Any]
    """All data should go into this (possibly nested) dict."""



import os
from typing import Dict, Optional, Any

from dotenv import dotenv_values

from confz.confz_source import ConfZEnvSource
from .loader import Loader


class EnvLoader(Loader):
    """Config loader for environment variables."""

    @classmethod
    def _transform_name(cls, name: str):
        return name.lower().replace("-", "_")

    @classmethod
    def _transform_remap(
        cls, map_in: Optional[Dict[str, str]]
    ) -> Optional[Dict[str, str]]:
        if map_in is None:
            return None

        map_out = {}
        for key, value in map_in.items():
            map_out[cls._transform_name(key)] = value
        return map_out

    @classmethod
    def _check_allowance(cls, var_name: str, confz_source: ConfZEnvSource) -> bool:
        if not confz_source.allow_all:
            if confz_source.allow is None:
                return False

            allow_list = [cls._transform_name(var) for var in confz_source.allow]
            if var_name not in allow_list:
                return False

        if confz_source.deny is not None:
            deny_list = [cls._transform_name(var) for var in confz_source.deny]
            if var_name in deny_list:
                return False

        return True

    @classmethod
    def populate_config(cls, config: dict, confz_source: ConfZEnvSource):
        remap = cls._transform_remap(confz_source.remap)

        origin_env_vars: Dict[str, Any] = dict(os.environ)
        if confz_source.file is not None:
            origin_env_vars = {**dotenv_values(confz_source.file), **origin_env_vars}

        env_vars = {}
        for env_var in origin_env_vars:
            var_name = env_var
            if confz_source.prefix is not None:
                if not var_name.startswith(confz_source.prefix):
                    continue
                var_name = var_name[len(confz_source.prefix) :]

            var_name = cls._transform_name(var_name)
            if not cls._check_allowance(var_name, confz_source):
                continue

            if remap is not None and var_name in remap:
                var_name = remap[var_name]

            env_vars[var_name] = origin_env_vars[env_var]

        env_vars = cls.transform_nested_dicts(
            env_vars, separator=confz_source.nested_separator
        )
        cls.update_dict_recursively(config, env_vars)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['confz', 'confz.loaders']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<7.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'python-dotenv>=0.19.2,<0.21.0']

setup_kwargs = {
    'name': 'confz',
    'version': '1.4.0',
    'description': 'ConfZ is a configuration management library for Python based on pydantic.',
    'long_description': '# ConfZ – Pydantic Config Management\n\n[![test](https://github.com/Zuehlke/ConfZ/actions/workflows/test.yml/badge.svg)](https://github.com/Zuehlke/ConfZ/actions/workflows/test.yml)\n[![documentation](https://readthedocs.org/projects/confz/badge/?version=latest)](https://confz.readthedocs.io/en/latest/)\n[![coverage](https://img.shields.io/badge/coverage-100%25-brightgreen)](https://github.com/Zuehlke/ConfZ/actions/workflows/coverage.yml)  <!-- hard-code because can not merge if below 100 -->\n[![python](https://img.shields.io/pypi/pyversions/confz)](https://pypi.org/project/confz/)\n[![pypi](https://img.shields.io/pypi/v/confz)](https://pypi.org/project/confz/)\n\n`ConfZ` is a configuration management library for Python based on [pydantic](https://pydantic-docs.helpmanual.io/).\nIt easily allows you to\n\n* load your configuration from config files, environment variables, command line arguments and more\n* transform the loaded data into a desired format and validate it\n* access the results as Python dataclass-like objects with full IDE support\n\nIt furthermore supports you in common use cases like:\n\n* Multiple environments\n* Singleton with lazy loading\n* Config changes for unit tests\n* Custom config sources\n\n\n## :package: Installation\n\n`ConfZ` is on [PyPI](https://pypi.org/project/confz/) and can be installed with pip:\n\n```shell\npip install confz\n```\n\n\n## :rocket: Quick Start\n\nThe first step of using `ConfZ` is to declare your config classes and sources, for example in `config.py`:\n\n```python\nfrom pathlib import Path\n\nfrom confz import ConfZ, ConfZFileSource\nfrom pydantic import SecretStr, AnyUrl\n\nclass DBConfig(ConfZ):\n    user: str\n    password: SecretStr\n\nclass APIConfig(ConfZ):\n    host: AnyUrl\n    port: int\n    db: DBConfig\n\n    CONFIG_SOURCES = ConfZFileSource(file=Path(\'/path/to/config.yml\'))\n```\n\nThanks to [pydantic](https://pydantic-docs.helpmanual.io/), you can use a wide variety of\n[field types](https://pydantic-docs.helpmanual.io/usage/types/) and\n[validators](https://pydantic-docs.helpmanual.io/usage/validators/).\n\nFrom now on, in any other file, you can access your config directly:\n\n```python\nfrom config import APIConfig\n\nprint(f"Serving API at {APIConfig().host}, port {APIConfig().port}.")\n```\n\nAs can be seen, the config does neither have to be loaded explicitly, nor instantiated globally. `ConfZ` automatically\nloads your config as defined in `CONFIG_SOURCES` the first time you access it. Thanks to its singleton mechanism, this\nhappens the first time only, afterwards you get back a cached,\n[immutable](https://pydantic-docs.helpmanual.io/usage/models/#faux-immutability) instance, behaving like any other\n_pydantic_ instance.\n\n```python\nassert APIConfig() is APIConfig()   # true because of singleton mechanism\nAPIConfig().port = 1234             # raises an error because of immutability\nAPIConfig().json()                  # call pydantic\'s method to get a json representation\n```\n\n**Note:** While the implicit and hidden loading of your config might be surprising and feel a bit like Python magic at\nfirst, it allows you to reduce a lot of boilerplate. Instead of having to load your config explicitly and then passing\nit down to all code layers that need it, you can directly access it from anywhere by just importing your config class\nand accessing for example `APIConfig().db.user` directly.\n\n### More Config Sources\n\n`ConfZ` is highly flexible in defining the source of your config. Do you have multiple environments? No Problem:\n\n```python\nfrom pathlib import Path\n\nfrom confz import ConfZ, ConfZFileSource\n\nclass MyConfig(ConfZ):\n    ...\n    CONFIG_SOURCES = ConfZFileSource(\n        folder=Path(\'/path/to/config/folder\'),\n        file_from_env=\'ENVIRONMENT\'\n    )\n```\n\nYour config file can now be defined in the environment variable `ENVIRONMENT` and is relative to `folder`.\n\nYou can also provide a list as config source and read for example from environment variables including a .env file and\nfrom command line arguments:\n\n```python\nfrom pathlib import Path\nfrom confz import ConfZ, ConfZEnvSource, ConfZCLArgSource\n\nclass MyConfig(ConfZ):\n    ...\n    CONFIG_SOURCES = [\n        ConfZEnvSource(allow_all=True, file=Path(".env.local")),\n        ConfZCLArgSource(prefix=\'conf_\')\n    ]\n```\n\n`ConfZ` now tries to populate your config either from environment variables having the same name as your attributes or\nby reading command line arguments that start with `conf_`. Recursive models are supported too, for example if you want\nto control the user-name in the API above, you can either set the environment variable `DB.USER` or pass the command\nline argument `--conf_db.user`.\n\n### Explicit Loading\n\nIn some scenarios, the config should not be a global singleton, but loaded explicitly and passed around locally.\nInstead of defining `CONFIG_SOURCES` as class variable, the sources can also be defined in the constructor directly:\n\n```python\nfrom pathlib import Path\n\nfrom confz import ConfZ, ConfZFileSource, ConfZEnvSource\n\nclass MyConfig(ConfZ):\n    number: int\n    text: str\n\nconfig1 = MyConfig(config_sources=ConfZFileSource(file=Path(\'/path/to/config.yml\')))\nconfig2 = MyConfig(config_sources=ConfZEnvSource(prefix=\'CONF_\', allow=[\'text\']), number=1)\nconfig3 = MyConfig(number=1, text=\'hello world\')\n```\n\nAs can be seen, additional keyword-arguments can be provided as well.\n\n**Note:** If neither class variable `CONFIG_SOURCES` nor constructor argument `config_sources` is provided, `ConfZ`\nbehaves like a regular _pydantic_ class.\n\n### Change Config Values\n\nIn some scenarios, you might want to change your config values, for example within a unit test. However, if you set the\n`CONFIG_SOURCES` class variable, this is not directly possible. To overcome this, every config class provides a context\nmanager to temporarily change your config:\n\n```python\nfrom pathlib import Path\n\nfrom confz import ConfZ, ConfZFileSource, ConfZDataSource\n\nclass MyConfig(ConfZ):\n    number: int\n    CONFIG_SOURCES = ConfZFileSource(file=Path(\'/path/to/config.yml\'))\n\nprint(MyConfig().number)                            # will print the value from the config-file\n\nnew_source = ConfZDataSource(data={\'number\': 42})\nwith MyConfig.change_config_sources(new_source):\n    print(MyConfig().number)                        # will print \'42\'\n\nprint(MyConfig().number)                            # will print the value from the config-file again\n```\n\n### Early Validation\n\nBy default, your config gets loaded the first time you instantiate the class, e.g. with `MyConfig().attribute`. This\nprevents side effects like loading a file while you import your config classes. If the config class cannot populate all\nmandatory fields in the correct format, _pydantic_ will raise an error at this point. To make sure this does not happen\nin an inconvenient moment, you can also instruct `ConfZ` to load all configs beforehand:\n\n```python\nfrom confz import validate_all_configs\n\nif __name__ == \'__main__\':\n    validate_all_configs()\n    # your application code\n```\n\nThe function `validate_all_configs` will instantiate all config classes defined in your code at any (reachable)\nlocation that have `CONFIG_SOURCES` set.\n\n\n## :book: Documentation\n\nNow you\'ve seen the two ways how `ConfZ` can be used: With class variable config sources, unlocking a singleton with\nlazy loading, or with keyword argument config sources, allowing to directly load your config values. In both cases,\ndefining your config sources from files, command line arguments and environment variables is highly flexible\n(and also extendable, by the way), while _pydantic_ still makes sure that everything matches your expectations in the\nend. You\'ve also seen how to temporarily change your config for example in unit tests and how to validate\nyour singleton config classes early in the code already.\n\nThe full documentation of `ConfZ`\'s features can be found at [readthedocs](https://confz.readthedocs.io/).\n\n\n## :information_source: About\n\n`ConfZ` was programmed and will be maintained by [Zühlke](https://www.zuehlke.com).\nThe first version was realized by [Silvan](https://github.com/silvanmelchior).\nSpecial thanks to Iwan with his [ConfMe](https://github.com/iwanbolzern/ConfMe), which inspired this project.\n\nWant to contribute to `ConfZ`? Check out the contribution [instruction & guidelines](CONTRIBUTING.md).\n',
    'author': 'Zühlke',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Zuehlke/ConfZ',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from .change import depends_on
from .confz import ConfZ
from .confz_source import (
    ConfZSources,
    ConfZSource,
    ConfZFileSource,
    ConfZEnvSource,
    ConfZCLArgSource,
    FileFormat,
    ConfZDataSource,
)
from .validate import validate_all_configs
